# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.

import itertools


def group_by_to_list(arr, key_fn):
    """
    Group the given list-like collection by the key computed using the given function. The
    collection does not have to be sorted in advance.

    @return a list of (key, list_of_values) tuples where keys are sorted

    >>> group_by_to_list([100, 201, 300, 301, 400], lambda x: x % 2)
    [(0, [100, 300, 400]), (1, [201, 301])]
    >>> group_by_to_list([100, 201, 300, 301, 400, 401, 402], lambda x: x % 3)
    [(0, [201, 300, 402]), (1, [100, 301, 400]), (2, [401])]

    """
    return [(k, list(v)) for (k, v) in itertools.groupby(sorted(arr, key=key_fn), key_fn)]


def group_by_to_dict(arr, key_fn):
    """
    Given a list-like collection and a function that computes a key, returns a map from keys to all
    values with that key.

    >>> group_by_to_dict([100, 201, 300, 301, 400], lambda x: x % 2)
    {0: [100, 300, 400], 1: [201, 301]}
    >>> group_by_to_dict([100, 201, 300, 301, 400, 401, 402], lambda x: x % 3)
    {0: [201, 300, 402], 1: [100, 301, 400], 2: [401]}
    """
    return dict(group_by_to_list(arr, key_fn))


def make_list(obj):
    """
    Convert the given object to a list. Strings get converted to a list of one string, not to a
    list of their characters. Sets are sorted.

    >>> make_list('asdf')
    ['asdf']
    >>> make_list(['a', 'b', 'c'])
    ['a', 'b', 'c']
    >>> make_list(set(['z', 'a', 'b']))
    ['a', 'b', 'z']
    >>> make_list(set(['z', 'a', 10, 20]))
    [10, 20, 'a', 'z']
    >>> make_list(set([10, 20, None, 'a', 'z']))
    [10, 20, None, 'a', 'z']
    """
    if isinstance(obj, str):
        return [obj]
    if isinstance(obj, set):
        # Sort by string representation because objects of different types are not comparable in
        # Python 3.
        return sorted(obj, key=lambda item: str(item))
    return list(obj)


def make_set(obj):
    if isinstance(obj, set):
        return obj
    return set(make_list(obj))



# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.

import logging


from yugabyte_pycommon.text_manipulation import get_bool_env_var


def is_verbose_mode():
    return get_bool_env_var('YB_VERBOSE')


def get_default_log_level():
    if is_verbose_mode():
        return logging.DEBUG
    return logging.INFO


def init_logging(log_level=None):
    if log_level is None:
        log_level = get_default_log_level()
    logging.basicConfig(
        level=log_level,
        format="[%(filename)s:%(lineno)d] %(asctime)s %(levelname)s: %(message)s"
    )



# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.
#

import os

from .base import TestCase

from yugabyte_pycommon import mkdir_p
from random import randint


class FileSystemUtilTestCase(TestCase):
    def test_mkdir_p(self):
        tmp_dir_path = '/tmp/yugabyte_mkdir_p_test_%d' % randint(1, 1000000000)
        for _ in range(2):
            mkdir_p(tmp_dir_path)
            self.assertTrue(os.path.isdir(tmp_dir_path))
        os.rmdir(tmp_dir_path)


# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase as PythonTestCase


class TestCase(PythonTestCase):
    pass



# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.
#

from yugabyte_pycommon import trim_long_text, run_program, quote_for_bash

from .base import TestCase


class TextManipulationsTestCase(TestCase):
    def test_trim_long_text(self):
        long_text = "\n".join([str(i) for i in range(1, 11)])
        self.assertEquals(long_text, trim_long_text(long_text, 20))
        self.assertEquals(long_text, trim_long_text(long_text, 10))
        self.assertEquals(
            "1\n2\n3\n4\n(2 lines skipped)\n7\n8\n9\n10", trim_long_text(long_text, 9))
        self.assertEquals(
            """1
2
(6 lines skipped)
9
10""", trim_long_text(long_text, 5))

        self.assertEquals('1\n2\n(6 lines skipped)\n9\n10', trim_long_text(long_text, 5))
        self.assertEquals("1\n(8 lines skipped)\n10", trim_long_text(long_text, 4))
        self.assertEquals("1\n(8 lines skipped)\n10", trim_long_text(long_text, 3))
        self.assertEquals("1\n(8 lines skipped)\n10", trim_long_text(long_text, 2))
        self.assertEquals("1\n(8 lines skipped)\n10", trim_long_text(long_text, 1))


    def test_quote_for_bash(self):
        for s in [
            '',
            'a',
            'a b'
            '"foo"'
            'foo"bar',
            'foo " bar',
            "foo'bar",
            '$foo $bar',
            "\\",
            "\\\\",
            "\\\\\\",
            '"' + "'",
            '"' + "'" + '"' + "'",
            r"""'\''""",
            'foo bar',
            ' ',
            'foo ',
            ' foo',
            ' foo bar '
            ';',
            ' ;',
            '; ',
            ' ; '
        ]:
            result = run_program('echo -n ' + quote_for_bash(s), shell=True)
            self.assertEquals(s, result.stdout)



# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.

"""Utilities for running external commands"""

import os
import sys
import logging
import subprocess
import threading

from yugabyte_pycommon.text_manipulation import cmd_line_args_to_str, decode_utf8, trim_long_text
from yugabyte_pycommon.logging_util import is_verbose_mode


# Default number of lines to shorten long stdout/stderr to.
DEFAULT_MAX_LINES_TO_SHOW = 1000


class ProgramResult:
    """
    This represents the result of executing an external program.
    """
    def __init__(self, cmd_line, cmd_line_str, returncode, stdout, stderr, program_path,
                 invocation_details_str, max_lines_to_show, output_captured):
        self.cmd_line = cmd_line
        self.cmd_line_str = cmd_line_str
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.program_path = program_path
        self.invocation_details_str = invocation_details_str
        self.max_lines_to_show = max_lines_to_show
        self.output_captured = output_captured

        self._set_error_msg()

    def success(self):
        return self.returncode == 0

    def failure(self):
        return self.returncode != 0

    def get_stdout_and_stderr_together(self):
        stdout_and_stderr = (
            self.get_user_friendly_stdout_msg() + self.get_user_friendly_stderr_msg())
        if not stdout_and_stderr:
            stdout_and_stderr = "No stdout or stderr from command: " + self.invocation_details_str
        return stdout_and_stderr

    def print_output_to_stdout(self):
        """
        Print both stdout and stderr of the external program to the stdout.
        """
        sys.stdout.write(self.get_stdout_and_stderr_together())
        sys.stdout.flush()

    def _set_error_msg(self):
        if self.returncode == 0:
            self.error_msg = None
            return

        self.error_msg = "Non-zero exit code {} from {}.".format(
            self.returncode,
            self.invocation_details_str,
            self.returncode, cmd_line_args_to_str)

        self.error_msg += self.get_user_friendly_stdout_msg()
        self.error_msg += self.get_user_friendly_stderr_msg()
        self.error_msg = self.error_msg.rstrip()

    def _wrap_for_error_msg(self, stream_type):
        assert stream_type in ['output', 'error']
        if stream_type == 'output':
            value = self.stdout
        else:
            value = self.stderr
        if value is None or not value.strip():
            return ""
        value = value.rstrip()
        return "\nStandard {} from {}:\n{}\n(end of standard {})\n".format(
            stream_type, self.invocation_details_str,
            trim_long_text(value, self.max_lines_to_show),
            stream_type)

    def get_user_friendly_stdout_msg(self):
        return self._wrap_for_error_msg("output")

    def get_user_friendly_stderr_msg(self):
        return self._wrap_for_error_msg("error")

    def raise_error_if_failed(self):
        if self.failure():
            raise ExternalProgramError(self.error_msg, self)


class ExternalProgramError(Exception):
    def __init__(self, message, result):
        self.message = message
        self.result = result


class WorkDirContext:
    """
    Allows setting a context for running external programs.

    Example:

    with WorkDirContext('/tmp'):
        run_program('ls')
    """
    def __init__(self, work_dir):
        self.thread_local = threading.local()
        self.work_dir = work_dir

    def __enter__(self):
        self.thread_local.old_dir = os.getcwd()
        os.chdir(self.work_dir)

    def __exit__(self, exception_type, exception_value, traceback):
        os.chdir(self.thread_local.old_dir)


def run_program(args, error_ok=False, report_errors=None, capture_output=True,
                max_lines_to_show=DEFAULT_MAX_LINES_TO_SHOW, cwd=None, shell=None, **kwargs):
    """
    Run the given program identified by its argument list, and return a ProgramResult object.
    :param error_ok: False to raise an exception on errors, True not to raise it.
    """
    if isinstance(args, str) and shell is None:
        shell = True

    if isinstance(args, str):
        # This is a special case, but very common.
        cmd_line_str = args
        args = [args]
    else:
        if isinstance(args, tuple):
            args = list(args)

        if isinstance(args, str):
            args = [args]

        def normalize_arg(arg):
            if isinstance(arg, int):
                return str(arg)
            return arg

        args = [normalize_arg(arg) for arg in args]

        cmd_line_str = cmd_line_args_to_str(args)

    invocation_details_str = "external program {{ %s }} running in '%s'" % (
            cmd_line_str, cwd or os.getcwd())

    if is_verbose_mode():
        logging.info("Running %s", invocation_details_str)

    try:
        output_redirection = subprocess.PIPE if capture_output else None
        program_subprocess = subprocess.Popen(
            args,
            stdout=output_redirection,
            stderr=output_redirection,
            shell=shell,
            cwd=cwd,
            **kwargs)
        program_stdout, program_stderr = program_subprocess.communicate()

    except OSError:
        logging.error("Failed to run %s", invocation_details_str)
        raise

    def cleanup_output(out_str):
        if out_str is None:
            return None
        return decode_utf8(out_str)

    clean_stdout = cleanup_output(program_stdout)
    clean_stderr = cleanup_output(program_stderr)

    result = ProgramResult(
        cmd_line=args,
        cmd_line_str=cmd_line_str,
        program_path=os.path.realpath(args[0]),
        returncode=program_subprocess.returncode,
        stdout=clean_stdout,
        stderr=clean_stderr,
        invocation_details_str=invocation_details_str,
        max_lines_to_show=max_lines_to_show,
        output_captured=capture_output)

    if program_subprocess.returncode != 0:
        if report_errors is None:
            report_errors = not error_ok
        if report_errors:
            logging.error(result.error_msg)
        if not error_ok:
            result.raise_error_if_failed()

    return result


def check_run_program(*args, **kwargs):
    """
    Similar to subprocess.check_call but using our run_program facility.
    """
    kwargs['capture_output'] = False
    kwargs['report_errors'] = True
    run_program(*args, **kwargs)
    return 0


def program_fails_no_log(args, **kwargs):
    """
    Run the given program, and returns if it failed. Does not log anything in case of success
    or failure.
    :param args: command line arguments or a single string to run as a shell command
    :param kwargs: additional keyword arguments for subprocess.Popen
    :return: True if the program succeeded
    """
    return run_program(args, error_ok=True, report_errors=False, **kwargs).failure()


def program_succeeds_no_log(args, **kwargs):
    """
    Run the given program, and returns True if it succeeded. Does not log anything in case of
    success or failure.
    :param args: command line arguments or a single string to run as a shell command
    :param kwargs: additional keyword arguments for subprocess.Popen
    :return: True if the program failed
    """
    return run_program(args, error_ok=True, report_errors=False, **kwargs).success()


def program_succeeds_empty_output(args, **kwargs):
    """
    Runs a program that is not expected to produce any output.
    :param args: command line arguments or a single string to run as a shell command
    :param kwargs: additional keyword arguments for subprocess.Popen
    :raises ExternalProgramError: if the program succeeds but produces extra output
    :return: True if the program succeeds and does not produce any output
    """
    result = run_program(args, error_ok=True, report_errors=False, **kwargs)
    if result.failure():
        return False

    if result.stdout.strip():
        error_msg = "Unexpected output in case of success. " + result.get_user_friendly_stdout_msg()
        logging.error(error_msg)
        raise ExternalProgramError(error_msg, result)

    return True



# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.
#

from .base import TestCase

import os

from yugabyte_pycommon import run_program, quote_for_bash, ExternalProgramError, WorkDirContext, \
    program_fails_no_log, program_succeeds_no_log, program_succeeds_empty_output, check_run_program
import logging
from testfixtures import LogCapture

class ExternalCommandsTestCase(TestCase):
    def test_run_program_noerror(self):
        result = run_program("true")
        self.assertEquals(0, result.returncode)
        self.assertEquals('', result.stdout)
        self.assertEquals('', result.stderr)

    def test_exit_codes(self):
        for exit_code in [0, 1, 2, 3, 10, 20, 100, 150, 200, 250, 255]:
            result = run_program("exit %d" % exit_code, shell=True, error_ok=exit_code != 0)
            self.assertEquals(exit_code, result.returncode)

    def test_error_reporing(self):
        with self.assertRaises(ExternalProgramError):
            run_program('false')

    def test_work_dir_context(self):
        old_work_dir = os.getcwd()
        for d in ['/tmp', os.path.expanduser('~')]:
            with WorkDirContext(d):
                self.assertEquals(d, os.getcwd())
                self.assertEquals(d, run_program('pwd').stdout.strip())

        self.assertEquals(old_work_dir, os.getcwd())

    def _capture_error_log_from_cmd(self, cmd):
        with LogCapture(level=logging.ERROR) as captured_logs:
            run_program(cmd, error_ok=True, report_errors=True)
        return str(captured_logs).strip()

    def test_log_error(self):
        self.assertRegexpMatches(
            self._capture_error_log_from_cmd("echo Output!; exit 1"), r"""
Non-zero exit code 1 from external program {{ echo Output!; exit 1 }} running in '.*'.
Standard output from external program {{ echo Output!; exit 1 }} running in '.*':
Output!
\(end of standard output\)""".strip())

        self.assertRegexpMatches(
            self._capture_error_log_from_cmd("echo Output!; echo Error! >&2; exit 1"), r"""
Non-zero exit code 1 from external program {{ echo Output!; echo Error! >&2; exit 1 }} running in '.*'.
Standard output from external program {{ echo Output!; echo Error! >&2; exit 1 }} running in '.*':
Output!
\(end of standard output\)
.*
Standard error from external program {{ echo Output!; echo Error! >&2; exit 1 }} running in '.*':
Error!
\(end of standard error\)""".strip())

    def test_shortcut_functions(self):
        self.assertTrue(program_fails_no_log('false'))
        self.assertFalse(program_fails_no_log('true'))
        self.assertFalse(program_succeeds_no_log('false'))
        self.assertTrue(program_succeeds_no_log('true'))
        self.assertTrue(program_succeeds_empty_output('true'))
        self.assertFalse(program_succeeds_empty_output('false'))
        with self.assertRaises(ExternalProgramError):
            self.assertTrue(program_succeeds_empty_output('echo foo'))

    def test_check_run_program(self):
        self.assertEquals(0, check_run_program('true'))
        with self.assertRaises(ExternalProgramError):
            check_run_program('false')

    def test_no_output_capture(self):
        result = run_program(
            'echo "This is expected to show up in the test output"',
            capture_output=False)
        self.assertIsNone(result.stdout)
        self.assertIsNone(result.stderr)
        self.assertFalse(result.output_captured)

    def test_error_msg(self):
        # No error message if exit code is 0.
        self.assertIsNone(run_program('true').error_msg)

        result = run_program(
            'echo " This is stdout! "; echo " This is stderr! " >&2; exit 1', error_ok=True)
        self.assertRegexpMatches(
            result.error_msg.strip(),
"""
Non-zero exit code 1 from external program {{ echo " This is stdout! "; echo " This is stderr! " >&2; exit 1 }} running in '.*'.
Standard output from external program {{ echo " This is stdout! "; echo " This is stderr! " >&2; exit 1 }} running in '.*':
 This is stdout!
\(end of standard output\)

Standard error from external program {{ echo " This is stdout! "; echo " This is stderr! " >&2; exit 1 }} running in '.*':
 This is stderr!
\(end of standard error\)
""".strip())

    def test_stdout_and_stderr_together(self):
        self.assertRegexpMatches(
            run_program('echo foo').get_stdout_and_stderr_together().strip(), r"""
Standard output from external program {{ echo foo }} running in '.*':
foo
\(end of standard output\)
            """.strip())
        self.assertRegexpMatches(
            run_program('echo bar >&2').get_stdout_and_stderr_together().strip(), r"""
Standard error from external program {{ echo bar >&2 }} running in '.*':
bar
\(end of standard error\)
            """.strip())
        self.assertRegexpMatches(
            run_program('echo foo; echo bar >&2').get_stdout_and_stderr_together().strip(), r"""
Standard output from external program {{ echo foo; echo bar >&2 }} running in '.*':
foo
\(end of standard output\)

Standard error from external program {{ echo foo; echo bar >&2 }} running in '.*':
bar
\(end of standard error\)
            """.strip())

